#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "petal/petal.h"
#include "timer.h"

#define PI M_PI

void openWindow(void);
void closeWindow(void);
void preCalculate(void);
void loadImage(void);
void setPalette(long long seed);
void doPlasma(void);

// info about the window size and depth
int xRes, yRes, bpp;

// the plasma sources
char *plasma1;
char *plasma2;

char *img;

Image *virtScreen;

// mappings for all the colours
long palette[256];

TIMER *timer;

int main(void) {
  openWindow();

  preCalculate();
  loadImage();
  doPlasma();
  closeWindow();
}

void openWindow(void) {
  char *name = "Plasma";

  if(init_petal("./petal/video.driver", name)) {
    printf("Cannot open window\n");
    exit(1);
  }

  if(set_resolution(320, 200, 32)) {
    printf("Cannot set resolution\n");
    exit(1);
  }

  get_resolution(&xRes, &yRes, &bpp);
  printf("Running at %dx%d @ %dbpp\n", xRes, yRes, bpp);

  plasma1 = new char[640 * 400];
  plasma2 = new char[640 * 400];

  img   = new char[320 * 200];

  virtScreen  = new_image();

  /*
  if(plasma1 == NULL || plasma2 == NULL ||
       image == NULL ||  virtScreen == NULL ||
       virtScreen->address == NULL) {
    printf("Couldn't allocate enough memory for screens\n");
    exit(1);
  }
  */
  set_image(virtScreen);
}

void closeWindow(void) {
  delete (plasma1);
  delete (plasma2);

  delete (img);

  delete_image(virtScreen);

  deinit_petal();
}

void preCalculate(void) {
  unsigned char *buf1 = (unsigned char *)plasma1;
  unsigned char *buf2 = (unsigned char *)plasma2;
  int i, j, dst = 0;

  for (j=0; j<400; j++) {
    for (i=0; i<640; i++) {
      buf1[dst] = (unsigned char)( 64 + 63 * ( sin( (double)hypot( 200-j, 320-i )/16 ) ) );
      buf2[dst] = (unsigned char)( 64 + 63 * sin( (float)i/(37+15*cos((float)j/74)) )
                                 * cos( (float)j/(31+11*sin((float)i/57))) );
      dst++;
    }
  }
}

void loadImage(void) {
  FILE *rawFile;

  rawFile = fopen( "image.raw", "rb" );
  fread(img, 1, 64000, rawFile );
  fclose(rawFile);
}

void setPalette(long long seed) {
  int i;
  //int r, g, b, c = 0;

  for (i=0; i<256; i++) {

    /*
    // I tried this because my colour routine wasn't work 100%, but
    // this gives the same result anyway... perhaps my video card
    // has a strange palette...!?
    r = (int)(128 + 127 * cos( i * PI / 128 + (double)seed/74 ));
    g = (int)(128 + 127 * sin( i * PI / 128 + (double)seed/63 ));
    b = (int)(128 - 127 * cos( i * PI / 128 + (double)seed/81 ));
    palette[i] = 0 + (r << 0) | (g << 8) | (b << 16);
    */


    palette[i] = colour( (32 + 31 * cos( i * PI / 128 + (double)seed/74 ))/64,
                         (32 + 31 * sin( i * PI / 128 + (double)seed/63 ))/64,
                         (32 - 31 * cos( i * PI / 128 + (double)seed/81 ))/64 );


    /*
    palette[i] = colour( 1 + cos( i * PI / 128 + (double)seed/74 )/2,
                         1 + sin( i * PI / 128 + (double)seed/63 )/2,
                         1 + cos( i * PI / 128 + (double)seed/81 )/2);
    */
  }
}

void doPlasma(void) {
  // define the plasma movement
  int x1, y1, x2, y2, x3, y3;
  long dst, i, j, src1, src2, src3;

  // start the timer
  timer = new TIMER;
  long long startTime = timer->getCount();
  long long frameCount = 0, currentTime;

  // run the main loop
  while (!get_key()) {
    // we get the time for this frame only once so that we are sure it stays
    // the same for all our calculations :)
    currentTime = timer->getCount() >> 14;

    // setup some nice colours, different every frame
    // this is a palette that wraps around itself, with different period sine
    // functions to prevent monotonous colours
    setPalette(currentTime);

    // move plasma with more sine functions :)
    x1 = 160 + (int)(159.0f * cos( (double) currentTime/ 97 ));
    x2 = 160 + (int)(159.0f * sin( (double)-currentTime/114 ));
    x3 = 160 + (int)(159.0f * sin( (double)-currentTime/137 ));
    y1 = 100 + (int)( 99.0f * sin( (double) currentTime/123 ));
    y2 = 100 + (int)( 99.0f * cos( (double)-currentTime/ 75 ));
    y3 = 100 + (int)( 99.0f * cos( (double)-currentTime/108 ));

    // we only select the part of the precalculated buffer that we need
    src1 = y1*640+x1;
    src2 = y2*640+x2;
    src3 = y3*640+x3;

    // draw the plasma... this is where most of the time is spent.
    // a good idea would be to add 4 pixels at a time by type casting
    // the unsigned chars to longs. I'll leave that to you :)
    dst = 0;
    for (j=0; j<200; j++) {
      for (i=0; i<320; i++) {
        // plot the pixel as a sum of all our plasma functions
	putpixel(i, j,
	          palette[(plasma1[src1] +
		           plasma2[src2] +
	                   plasma2[src3] +
			       img[dst]) & 0xFF
	          ],
		  virtScreen);
        dst++; src1++; src2++; src3++;
      }

      // get the next line in the precalculated buffers
      src1 += 320; src2 += 320; src3 += 320;
    }

    // dump our temporary buffer to the screen
    //vga->Update();
    show_image(virtScreen);
    frameCount++;
  }

  // calculate total running time
  long long totalTime = timer->getCount() - startTime;

  // and turn off the timer
  delete (timer);

  // now get FPS
  float FPS = (float)(frameCount) / (float)(totalTime * CLOCKS_PER_SEC);

  /*
  // return to text mode
  delete (vga);

  // free memory
  delete [] (image);
  delete [] (plasma1);
  delete [] (plasma2);
  */

  // make sure every one knows about flipcode :)
  cout << "                             The Art Of" << endl;
  cout << "                         D E M O M A K I N G " << endl << endl;
  cout << "                       by Alex J. Champandard" << endl << endl;
  cout << "       Go to http://www.flipcode.com/demomaking for more information." << endl;
  cout << endl << FPS << " frames per second." << endl;

  // we've finished!
  return;
}